import { ValidationError } from "../Errors";
import { RawSchema } from "../schema/raw/Schema";
import {
  TransactionResult,
  Transaction,
  isTransactionQueryRef,
} from "./Transaction";

/**
 * Type representing an RQL request and its associated parameters.
 */
export type RQLQuery = [string, RQLParams];

export type RQLParamValue = string | number | boolean | null;
/**
 * Type representing RQL parameters to send along a request.
 */
export interface RQLParams {
  [key: string]: RQLParamValue;
}

/**
 * Type representing a row returned in a result set.
 */
export type RqlRow = Array<string | number | boolean | null>;

/**
 * Type representing a RQL request result.
 */
export type ResultSet = Array<RqlRow>;

/**
 * Type representing the currently logged user's data
 */
export interface CurrentUser {
  login: string;
  eid: number;
  dcTitle: string;
}

/**
 * Type representing the body of an error sent by CubicWeb's API.
 */
type ApiError = {
  title: string;
  message: string;
  data?: Record<string, unknown>;
};

/**
 * Type representing the error data sent to clients when retrieving an API error.
 */
export type ApiErrorResponse = {
  status: number;
} & ApiError;

/**
 * Class used to handle requests to the CubicWeb API.
 *
 * You should never have to create this object yourself.
 * Instead, retrieve this object from the methods in the `Client` class.
 *
 * @internal
 */
export default class Api {
  private _apiUrl: string;
  private _version = "v1";
  private _rqlApiUrl: string;
  private _schemaApiUrl: string;
  private _loginApiUrl: string;
  private _currentUserApiUrl: string;
  private _transactionApiUrl: string;

  /**
   * Class used to handle requests to the CubicWeb API.
   *
   * You should never have to create this object yourself.
   * Instead, retrieve this object from the methods in the `Client` class.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(apiUrl: string) {
    this._apiUrl = apiUrl;
    this._rqlApiUrl = this.getApiRoute("rql");
    this._schemaApiUrl = this.getApiRoute("schema");
    this._loginApiUrl = this.getApiRoute("login");
    this._currentUserApiUrl = this.getApiRoute("current-user");
    this._transactionApiUrl = this.getApiRoute("transaction");
  }

  get apiUrl() {
    return this._apiUrl;
  }

  private getApiRoute(route: string) {
    return `${this._apiUrl}/${this._version}/${route}`;
  }

  /**
   * Executes the given query and waits for the response.
   *
   * @param query The RQL query to send
   * @param params The additional parameters for the request
   * @returns A promise with the request's result set
   */
  async execute(query: string, params: RQLParams): Promise<ResultSet> {
    try {
      return await nonNullFetchApi<ResultSet>(this._rqlApiUrl, {
        method: "POST",
        body: JSON.stringify({ query, params }),
        credentials: "include", // FIXME is it secure?
      });
    } catch (e) {
      this.handleUserErrors(e);
    }
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema(): Promise<RawSchema> {
    return nonNullFetchApi<RawSchema>(this._schemaApiUrl, {
      method: "GET",
      credentials: "include",
    });
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password THe user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string): Promise<boolean> {
    try {
      await fetchApi<null>(this._loginApiUrl, {
        method: "POST",
        body: JSON.stringify({
          login,
          password,
        }),
        credentials: "include",
      });
      return true;
    } catch (_error) {
      return false;
    }
  }

  /**
   * Get the current user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await nonNullFetchApi<CurrentUser>(this._currentUserApiUrl, {
      method: "GET",
      credentials: "include",
    });
  }

  async executeTransaction(
    transaction: Transaction
  ): Promise<TransactionResult> {
    const data = transaction.queries.map((transactionQuery) => ({
      query: transactionQuery.query,
      params: Object.fromEntries(
        Object.entries(transactionQuery.params).map(([key, value]) => [
          key,
          isTransactionQueryRef(value)
            ? {
                queryIndex: value.queryIndex,
                row: value.row,
                column: value.column,
              }
            : value,
        ])
      ),
    }));
    try {
      const resultSets = await nonNullFetchApi<ReadonlyArray<ResultSet>>(
        this._transactionApiUrl,
        {
          method: "POST",
          credentials: "include",
          body: JSON.stringify(data),
        }
      );
      return new TransactionResult(resultSets);
    } catch (e) {
      this.handleUserErrors(e);
    }
  }

  private handleUserErrors(error: unknown): never {
    if (
      typeof error === "object" &&
      error !== null &&
      "title" in error &&
      typeof error.title === "string" &&
      "message" in error &&
      typeof error.message === "string"
    ) {
      if (error.title === "ValidationError") {
        throw new ValidationError(error.message);
      }
      throw error;
    }
    throw error;
  }
}

/**
 * Behaves the same as `fetchAPI` but forbids null responses.
 * If this function detects a null response from the API,
 * the promise will be rejected with a 500 error code.
 *
 * @param input The API route to use
 * @param init Additional request parameters
 * @returns A promise with the non-null fetched result
 */
async function nonNullFetchApi<R = unknown>(
  input: RequestInfo,
  init?: RequestInit
) {
  const result = await fetchApi<R>(input, init);
  if (result == null) {
    const error: ApiErrorResponse = {
      status: 500,
      title: "Unexpected null response",
      message: "Response was null but a value was expected",
    };
    throw error;
  }
  return result;
}

/**
 * Sends a request to the given API route with the `Content-Type` header set and waits for a response.
 * The returned promise will be rejected if the API returns HTTP error codes.
 * The rejected promise will contain an `ApiErrorResponse` object.
 *
 * @param input The API route to use
 * @param init Additional request parameters
 * @returns A promise with the fetched result
 */
async function fetchApi<R = unknown>(
  input: RequestInfo,
  init?: RequestInit
): Promise<R | null> {
  let headers = init?.headers;
  if (init?.method && init.method !== "GET" && init.method !== "DELETE") {
    headers = {
      "Content-Type": "application/json;charset=UTF-8",
      ...init?.headers,
    };
  }

  const response = await fetch(input, {
    ...init,
    headers,
  });
  const status = response.status;
  if (status >= 200 && status < 300) {
    if (status === 204) {
      return null;
    }
    try {
      return await response.json();
    } catch (e) {
      console.error(e);
      return null;
    }
  } else {
    let json;
    try {
      json = await response.json();
    } catch (e) {
      console.error(e);
      throw { status, title: "Error", message: "" };
    }
    throw { status, ...json };
  }
}
