import { EntityRole } from "../classes/Utils";
import { AttributeConstraint, RelationConstraint } from "./Constraints";
import { EntityRawSchema, EntityRawSchemaArray } from "./Entities";

/**
 * Type representing all the possible cardinalities in CubicWeb.
 */
export type Cardinality = "*" | "1" | "?" | "+";

/**
 * Type representing a cardinality pair as used in relation definitions.
 */
export type CardinalityPair = `${Cardinality}${Cardinality}`;

/**
 * Type used as a base to build final relation definitions with primitive types (Int, String, ...).
 *
 * @internal
 */
type BuildObjBase<T extends string> = {
  object: T;
  final: true;
  required?: boolean;
};

interface StringBuildObj extends BuildObjBase<"String"> {
  default: string | null;
}

interface PasswordBuildObj extends BuildObjBase<"Password"> {
  default: string | null;
}

interface BytesBuildObj extends BuildObjBase<"Bytes"> {
  default: string | null;
}

interface IntBuildObj extends BuildObjBase<"Int"> {
  default: number | null;
}

interface BigIntBuildObj extends BuildObjBase<"BigInt"> {
  default: number | null;
}

interface FloatBuildObj extends BuildObjBase<"Float"> {
  default: number | null;
}

interface BooleanBuildObj extends BuildObjBase<"Boolean"> {
  default: boolean | null;
}

interface DecimalBuildObj extends BuildObjBase<"Decimal"> {
  default: number | null;
}

interface DateBuildObj extends BuildObjBase<"Date"> {
  default: unknown;
}

interface TimeBuildObj extends BuildObjBase<"Time"> {
  default: unknown;
}

interface DatetimeBuildObj extends BuildObjBase<"Datetime"> {
  default: Date | "NOW" | "TODAY" | null;
}

interface TZTimeBuildObj extends BuildObjBase<"TZTime"> {
  default: unknown;
}

interface TZDatetimeBuildObj extends BuildObjBase<"TZDatetime"> {
  default: unknown;
}

interface IntervalBuildObj extends BuildObjBase<"Interval"> {
  default: unknown;
}

interface CustomBuildObj<T extends string> extends BuildObjBase<T> {
  type: T;
  default: unknown;
}

/**
 * Type allowing to link the type of an object to the type of their default values in a
 * relation definition. It only works for final relations using primitive types.
 *
 * @internal
 */
export type BuildObj = Readonly<
  | StringBuildObj
  | PasswordBuildObj
  | BytesBuildObj
  | IntBuildObj
  | BigIntBuildObj
  | FloatBuildObj
  | BooleanBuildObj
  | DecimalBuildObj
  | DateBuildObj
  | TimeBuildObj
  | DatetimeBuildObj
  | TZTimeBuildObj
  | TZDatetimeBuildObj
  | IntervalBuildObj
  | CustomBuildObj<string>
>;

/**
 * Type representing the possible options for an attribute.
 */
export type AttributeOptions = {
  indexed?: boolean;
  fulltextindexed?: boolean;
  internationalizable?: boolean;
};

/**
 * Type representing the possible options for a relation definition.
 */
export type RelationOptions = {
  symmetric?: boolean;
  inlined?: boolean;
  fulltext_container?: EntityRole;
  composite?: EntityRole;
};

/**
 * Type representing a relation definition schema.
 * It can represent both attributes and relations depending on the value of Final.
 *
 * It can infer default values type depending on the object type if the relation is final.
 */
export type RelationDefinitionRawSchema<
  ESchemas extends EntityRawSchemaArray,
  Final extends boolean = boolean
> = Readonly<
  {
    type: string;
    description: string;
    cardinality: CardinalityPair;
    subject: ESchemas[number]["type"];
  } & (Final extends true
    ? BuildObj & {
        final: true;
        constraints: AttributeConstraint[];
        options: AttributeOptions;
      }
    : {
        object: EntityRawSchema["type"];
        final: false;
        constraints: RelationConstraint[];
        options: RelationOptions;
      })
>;

/**
 * Type representing a list of relation definitions.
 */
export type RelationDefinitionRawSchemaArray<
  ESchemas extends EntityRawSchemaArray
> = ReadonlyArray<RelationDefinitionRawSchema<ESchemas, boolean>>;
