import {
  EntitySchema,
  getLooseSchema,
  Schema,
  Transaction,
} from "@cubicweb/client";
import Api from "@cubicweb/client/lib/api/Api";
import {
  generateRqlCreateRelationsParts,
  GetEntitiesToCreateFromInput,
  getChildEntitiesToCreate,
  createNewEntities,
  mergeNewElementsIntoData,
  generateRqlCreateParts,
} from "../..//utils/createParams";
import * as createParams from "../../utils/createParams";
import { initVariableKeyAssociation } from "../../utils/constants";

const schema = getLooseSchema(
  new Schema({
    entities: [
      {
        type: "Person",
        description: "A Person",
        final: false,
      },
      {
        type: "Task",
        description: "Something that must be done",
        final: false,
      },
      {
        type: "String",
        description: "",
        final: true,
      },
    ],
    relations_definitions: [
      {
        type: "has_task",
        description: "Something that the person must do",
        final: false,
        cardinality: "*+",
        subject: "Person",
        object: "Task",
        constraints: [],
        options: {},
      },
      {
        type: "name",
        description: "",
        final: true,
        cardinality: "11",
        subject: "Person",
        object: "String",
        constraints: [],
        default: null,
        options: {},
      },
    ],
  })
);

describe("getCreateRelationRqlParams", () => {
  const relationsMatched = schema.matchRelationDefinitions({
    subject: "Person",
    type: "has_task",
    object: "Task",
  });
  const relation = relationsMatched[0];
  describe("subject", () => {
    it("should create one parameter", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "subject",
          [1]
        )
      ).toEqual({
        assignments: ["X has_task REL_10"],
        restrictions: ["REL_10 eid 1"],
      });
    });
    it("should create two parameters", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "subject",
          [1, 2]
        )
      ).toEqual({
        assignments: ["X has_task REL_10", "X has_task REL_11"],
        restrictions: ["REL_10 eid 1", "REL_11 eid 2"],
      });
    });
    it("should not create parameters", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "subject",
          []
        )
      ).toEqual({
        assignments: [],
        restrictions: [],
      });
    });
  });

  describe("object", () => {
    it("should create one parameter", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "object",
          [1]
        )
      ).toEqual({
        assignments: ["REL_10 has_task X"],
        restrictions: ["REL_10 eid 1"],
      });
    });
    it("should create two parameters", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "object",
          [1, 2]
        )
      ).toEqual({
        assignments: ["REL_10 has_task X", "REL_11 has_task X"],
        restrictions: ["REL_10 eid 1", "REL_11 eid 2"],
      });
    });
    it("should not create parameters", () => {
      expect(
        generateRqlCreateRelationsParts(
          relation,
          initVariableKeyAssociation(),
          "object",
          []
        )
      ).toEqual({
        assignments: [],
        restrictions: [],
      });
    });
  });
});

describe("GetEntitiesToCreateFromInput", () => {
  it("should detect when there is nothing to create", () => {
    expect(
      GetEntitiesToCreateFromInput({ selected: [1, 2], id: "temp_input" })
    ).toEqual([]);
    expect(GetEntitiesToCreateFromInput(1)).toEqual([]);
    expect(GetEntitiesToCreateFromInput(undefined)).toEqual([]);
  });
  it("should find single elements", () => {
    expect(GetEntitiesToCreateFromInput({})).toEqual([{}]);
    expect(GetEntitiesToCreateFromInput({ test: 1 })).toEqual([{ test: 1 }]);
  });
  it("should find objects to create", () => {
    expect(
      GetEntitiesToCreateFromInput({
        created: [
          {},
          { id: 12 },
          { chose: "chose" },
          { id: 15, machin: "truc" },
        ],
        id: "temp_input",
      })
    ).toEqual([{}, { id: 12 }, { chose: "chose" }, { id: 15, machin: "truc" }]);
    expect(
      GetEntitiesToCreateFromInput({
        created: [
          {},
          { id: 12 },
          { chose: "chose" },
          { id: 15, machin: "truc" },
        ],
        selected: [1],
        id: "temp_input",
      })
    ).toEqual([{}, { id: 12 }, { chose: "chose" }, { id: 15, machin: "truc" }]);
  });
});

describe("getChildEntitiesToCreate", () => {
  const personEntitySchema = schema.getEntity("Person");
  it("should return an empty array", () => {
    expect(getChildEntitiesToCreate(personEntitySchema, { test: [] })).toEqual(
      []
    );
    expect(
      getChildEntitiesToCreate(personEntitySchema, {
        has_task: { selected: [], id: "temp_input" },
      })
    ).toEqual([]);
    expect(
      getChildEntitiesToCreate(personEntitySchema, {
        has_task: { removed: [1, 2], id: "temp_input" },
      })
    ).toEqual([]);
  });

  it("should ignore invalid relations", () => {
    expect(
      getChildEntitiesToCreate(personEntitySchema, { test: [{ thing: 1 }] })
    ).toEqual([]);
    expect(
      getChildEntitiesToCreate(personEntitySchema, {
        reverse_has_task: [{ thing: 1 }],
      })
    ).toEqual([]);
  });

  it("should handle regular relations", () => {
    expect(
      getChildEntitiesToCreate(personEntitySchema, {
        Person__has_task__Task: {
          created: [
            {
              name: "Test",
            },
          ],
          id: "temp_input",
        },
      })
    ).toEqual([
      {
        type: "Task",
        relation: "Person__has_task__Task",
        data: {
          name: "Test",
        },
      },
    ]);
    expect(
      getChildEntitiesToCreate(personEntitySchema, {
        Person__has_task__Task: {
          created: [
            {
              name: "Test",
            },
          ],
          selected: [1, 2],
          id: "temp_input",
        },
      })
    ).toEqual([
      {
        type: "Task",
        relation: "Person__has_task__Task",
        data: {
          name: "Test",
        },
      },
    ]);
  });

  const taskEntitySchema = schema.getEntity("Task");
  it("should handle reverse relations", () => {
    expect(
      getChildEntitiesToCreate(taskEntitySchema, {
        reverse_Person__has_task__Task: {
          created: [
            {
              name: "Test",
            },
          ],
          id: "temp_input",
        },
      })
    ).toEqual([
      {
        type: "Person",
        relation: "reverse_Person__has_task__Task",
        data: {
          name: "Test",
        },
      },
    ]);
  });
});

describe("createNewEntities", () => {
  it("should return an empty array", () => {
    const transaction = new Transaction();
    expect(createNewEntities(schema, transaction, [])).toEqual({});
  });

  it("should return one item", () => {
    const transaction = new Transaction();
    expect(
      createParams.createNewEntities(schema, transaction, [
        {
          type: "Task",
          relation: "has_task",
          data: {
            name: "Test",
          },
        },
      ])
    ).toEqual({
      has_task: [
        {
          queryIndex: 0,
          row: 0,
          column: 0,
        },
      ],
    });
  });

  it("should return two items for the same relation", async () => {
    const transaction = new Transaction();
    expect(
      createParams.createNewEntities(schema, transaction, [
        {
          type: "Task",
          relation: "has_task",
          data: {
            name: "Test",
          },
        },
        {
          type: "Task",
          relation: "has_task",
          data: {
            name: "Test 2",
          },
        },
      ])
    ).toEqual({
      has_task: [
        {
          queryIndex: 0,
          row: 0,
          column: 0,
        },
        {
          queryIndex: 1,
          row: 0,
          column: 0,
        },
      ],
    });
  });
  it("should return one item for each relation", async () => {
    const transaction = new Transaction();
    expect(
      createParams.createNewEntities(schema, transaction, [
        {
          type: "Task",
          relation: "has_task",
          data: {
            name: "Test",
          },
        },
        {
          type: "Task",
          relation: "test",
          data: {
            name: "Test 2",
          },
        },
      ])
    ).toEqual({
      has_task: [
        {
          queryIndex: 0,
          row: 0,
          column: 0,
        },
      ],
      test: [
        {
          queryIndex: 1,
          row: 0,
          column: 0,
        },
      ],
    });
  });
});

describe("mergeNewElementsIntoData", () => {
  it("should return an empty object", () => {
    expect(mergeNewElementsIntoData({}, {})).toEqual({});
  });

  it("should return the same object", () => {
    expect(
      mergeNewElementsIntoData(
        {
          has_task: [1],
        },
        {}
      )
    ).toEqual({ has_task: [1] });
  });
  it("should ignore new keys", () => {
    expect(
      mergeNewElementsIntoData(
        {},
        {
          has_task: [
            {
              queryIndex: 0,
              column: 0,
              row: 0,
            },
          ],
        }
      )
    ).toEqual({});
  });
  it("should merge new element", () => {
    expect(
      mergeNewElementsIntoData(
        {
          has_task: {
            created: [{ name: "test" }],
            selected: [1],
            id: "temp_input",
          },
        },
        {
          has_task: [
            {
              queryIndex: 0,
              column: 0,
              row: 0,
            },
          ],
        }
      )
    ).toEqual({
      has_task: {
        created: [
          {
            name: "test",
          },
        ],
        selected: [
          1,
          {
            queryIndex: 0,
            column: 0,
            row: 0,
          },
        ],
        id: "temp_input",
      },
    });
    expect(
      mergeNewElementsIntoData(
        {
          has_task: {
            created: [{ name: "test" }],
            selected: [1],
            id: "temp_input",
          },
          test: {
            selected: [1],
            id: "temp_input",
          },
          test2: {
            created: [{ name: "test" }],
            id: "temp_input",
          },
        },
        {
          has_task: [
            {
              queryIndex: 0,
              column: 0,
              row: 0,
            },
          ],
          test2: [
            {
              queryIndex: 1,
              column: 0,
              row: 0,
            },
          ],
        }
      )
    ).toEqual({
      has_task: {
        created: [
          {
            name: "test",
          },
        ],
        selected: [
          1,
          {
            queryIndex: 0,
            column: 0,
            row: 0,
          },
        ],
        id: "temp_input",
      },
      test: {
        selected: [1],
        id: "temp_input",
      },
      test2: {
        created: [{ name: "test" }],
        selected: [
          {
            queryIndex: 1,
            column: 0,
            row: 0,
          },
        ],
        id: "temp_input",
      },
    });
  });
});

describe("findAndCreateChildEntities", () => {
  const personEntitySchema = schema.getEntity("Person") as EntitySchema;
  it("handles when there is no data", () => {
    const transaction = new Transaction();
    expect(
      createParams.findAndCreateChildEntities(
        schema,
        personEntitySchema,
        transaction,
        {}
      )
    ).toEqual({});
  });
  it("handles when there is no new data", () => {
    const transaction = new Transaction();
    expect(
      createParams.findAndCreateChildEntities(
        schema,
        personEntitySchema,
        transaction,
        { Person__has_task__Task: { id: "temp_input", selected: [5] } }
      )
    ).toEqual({
      Person__has_task__Task: { id: "temp_input", selected: [5] },
    });
  });
  it("handles when there is new data", () => {
    const transaction = new Transaction();
    expect(
      createParams.findAndCreateChildEntities(
        schema,
        personEntitySchema,
        transaction,
        {
          Person__has_task__Task: {
            id: "temp_input",
            created: [{ name: "test" }],
          },
        }
      )
    ).toEqual({
      Person__has_task__Task: {
        id: "temp_input",
        created: [{ name: "test" }],
        selected: [
          {
            queryIndex: 0,
            row: 0,
            column: 0,
          },
        ],
      },
    });
  });
});

describe("generateRqlCreateParts", () => {
  const entity = schema.getEntity("Person") as EntitySchema;
  it("should return empty assignements", () => {
    expect(
      generateRqlCreateParts(entity, {}, initVariableKeyAssociation())
    ).toEqual({
      assignments: [],
      restrictions: [],
    });
  });
  it("should have one attribute assignement", () => {
    expect(
      generateRqlCreateParts(
        entity,
        { name: "test" },
        initVariableKeyAssociation()
      )
    ).toEqual({
      assignments: ["X name %(name)s"],
      restrictions: [],
    });
  });
  it("should have one relation assignement", () => {
    expect(
      generateRqlCreateParts(
        entity,
        {
          Person__has_task__Task: { selected: [1], id: "temp_input" },
        },
        initVariableKeyAssociation()
      )
    ).toEqual({
      assignments: ["X has_task REL_10"],
      restrictions: ["REL_10 eid 1"],
    });
  });
  it("should have one attribute and one relation assignements", () => {
    expect(
      generateRqlCreateParts(
        entity,
        {
          name: "test",
          Person__has_task__Task: { selected: [1], id: "temp_input" },
        },
        initVariableKeyAssociation()
      )
    ).toEqual({
      assignments: ["X name %(name)s", "X has_task REL_10"],
      restrictions: ["REL_10 eid 1"],
    });
  });
});

describe("makeRecursiveCreateRequest", () => {
  it("detects invalid entities", () => {
    const transaction = new Transaction();
    expect(() =>
      createParams.makeRecursiveCreateRequest(
        schema,
        "Invalid",
        {},
        transaction
      )
    ).toThrow();
  });
  it("handles empty form data", () => {
    const transaction = new Transaction();
    expect(
      createParams.makeRecursiveCreateRequest(schema, "Person", {}, transaction)
    ).toEqual({ queryIndex: 0, row: 0, column: 0 });
  });
  // TODO return an edited transaction object instead of the added query
});
