import {
  EntityRawSchema,
  EntityRawSchemaArray,
  RelationDefinitionRawSchema,
  Schema,
} from "@cubicweb/client";
import { getChangedData, rowToRecord } from "../../utils/dataset";

const personEntity: EntityRawSchema = {
  type: "Person",
  description: "A Person",
  final: false,
};

const taskEntity: EntityRawSchema = {
  type: "Task",
  description: "Something that must be done",
  final: false,
};

const hasMainTaskRelation: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "has_main_task",
  description: "Something that the person must do",
  final: false,
  cardinality: "11",
  subject: "Person",
  object: "Task",
  constraints: [],
  options: {},
};

const schema = new Schema<EntityRawSchemaArray>({
  entities: [personEntity, taskEntity],
  relations_definitions: [hasMainTaskRelation],
});

describe("rowToRecord", () => {
  const identity = <E>(x: E) => x;
  it("should return an empty object", () => {
    expect(rowToRecord([], [], [])).toMatchObject({});
  });
  describe("handling ids", () => {
    it("should only have an id", () => {
      expect(rowToRecord([0], ["id"], [identity])).toMatchObject({
        eid: 0,
        id: 0,
      });
    });
    it("should convert string ids", () => {
      expect(rowToRecord(["0"], ["id"], [identity])).toMatchObject({
        eid: 0,
        id: 0,
      });
    });
    it("should convert string ids", () => {
      expect(rowToRecord(["0"], ["id"], [identity])).toMatchObject({
        eid: 0,
        id: 0,
      });
    });
    it("should use raw string id", () => {
      expect(rowToRecord(["test"], ["id"], [identity])).toMatchObject({
        eid: "test",
        id: "test",
      });
    });
  });
  describe("handling data", () => {
    it("should use raw data", () => {
      expect(
        rowToRecord(
          [0, 1, "machin", false, null],
          ["id", "test1", "test2", "test3", "test4"],
          [identity, identity, identity, identity, identity]
        )
      ).toMatchObject({
        eid: 0,
        id: 0,
        test1: 1,
        test2: "machin",
        test3: false,
        test4: null,
      });
    });
    it("should use selection transformers", () => {
      expect(
        rowToRecord(
          [0, 1, "machin", false, null],
          ["id", "test1", "test2", "test3", "test4"],
          [
            identity,
            (x) => (x as number) + 15,
            (x) => x + " chose",
            (x) => !x,
            () => "test",
          ]
        )
      ).toMatchObject({
        eid: 0,
        id: 0,
        test1: 16,
        test2: "machin chose",
        test3: true,
        test4: "test",
      });
    });
    it("should detect arrays of different lengths", () => {
      expect(() => rowToRecord([0, 1], ["id"], [identity])).toThrowError(
        "Arrays are of different lengths"
      );
      expect(() => rowToRecord([0], ["id", "test"], [identity])).toThrowError(
        "Arrays are of different lengths"
      );
      expect(() => rowToRecord([0], ["id"], [identity, identity])).toThrowError(
        "Arrays are of different lengths"
      );
    });
  });
});

describe("getChangedData", () => {
  it("should detect same attributes", () => {
    expect(getChangedData(schema, {}, {})).toMatchObject({});
    expect(getChangedData(schema, { test1: 1 }, { test1: 1 })).toMatchObject(
      {}
    );
    expect(
      getChangedData(
        schema,
        { test1: 1, test2: "test" },
        { test1: 1, test2: "test" }
      )
    ).toMatchObject({});
  });
  it("should find different attributes", () => {
    expect(getChangedData(schema, { test1: 1 }, { test1: 2 })).toMatchObject({
      changedData: { test1: 1 },
      removedData: {},
    });
    expect(
      getChangedData(
        schema,
        { test1: 1, test2: "test" },
        { test1: 2, test2: "testy" }
      )
    ).toMatchObject({
      changedData: { test1: 1, test2: "test" },
      removedData: {},
    });
  });
  it("should find only different attribute", () => {
    expect(
      getChangedData(schema, { test1: 1, test2: 2 }, { test1: 2, test2: 2 })
    ).toMatchObject({
      changedData: { test1: 1 },
      removedData: {},
    });
  });
  it("should find new items in relation arrays", () => {
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: [1, 2] },
        { Person__has_main_task__Task: [1] }
      )
    ).toMatchObject({
      changedData: { Person__has_main_task__Task: [2] },
      removedData: {},
    });
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: [1, 2], test2: [1, 2] },
        { Person__has_main_task__Task: [1], test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: { Person__has_main_task__Task: [2] },
      removedData: {},
    });
  });
  it("should find different relation value", () => {
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: 1, test2: [1, 2] },
        { Person__has_main_task__Task: 2, test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: { Person__has_main_task__Task: 1 },
      removedData: {},
    });
  });
  it("should find new attribute values", () => {
    expect(
      getChangedData(schema, { test1: 1, test2: 2 }, { test2: 2 })
    ).toMatchObject({
      changedData: { test1: 1 },
      removedData: {},
    });
  });
  it("should find new relation array values", () => {
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: [1], test2: 2 },
        { test2: 2 }
      )
    ).toMatchObject({
      changedData: { Person__has_main_task__Task: [1] },
      removedData: {},
    });
  });
  it("should find new relation value", () => {
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: 1, test2: [1, 2] },
        { test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: { Person__has_main_task__Task: 1 },
      removedData: {},
    });
  });
  it("should find deleted attributes", () => {
    expect(getChangedData(schema, {}, { test2: 2 })).toMatchObject({
      changedData: { test2: undefined },
      removedData: {},
    });
    expect(getChangedData(schema, { test1: 1 }, { test2: 2 })).toMatchObject({
      changedData: { test1: 1, test2: undefined },
      removedData: {},
    });
  });
  it("should find deleted items in relation arrays", () => {
    expect(
      getChangedData(
        schema,
        { test1: [1], test2: [1, 2] },
        { test1: [1, 2], test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: {},
      removedData: { test1: [2] },
    });
  });
  it("should find deleted relation array", () => {
    expect(
      getChangedData(
        schema,
        { test2: [1, 2] },
        { Person__has_main_task__Task: [1, 2], test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: {},
      removedData: { Person__has_main_task__Task: [1, 2] },
    });
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: undefined, test2: [1, 2] },
        { Person__has_main_task__Task: [1, 2], test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: {},
      removedData: { Person__has_main_task__Task: [1, 2] },
    });
  });
  it("should find deleted relation", () => {
    expect(
      getChangedData(
        schema,
        { Person__has_main_task__Task: undefined, test2: [1, 2] },
        { Person__has_main_task__Task: 1, test2: [1, 2] }
      )
    ).toMatchObject({
      changedData: {},
      removedData: { Person__has_main_task__Task: 1 },
    });
  });
});
