import {
  EntityRawSchema,
  EntityRawSchemaArray,
  EntitySchema,
  RelationDefinitionRawSchema,
  Schema,
} from "@cubicweb/client";
import {
  getAllRqlParts,
  getAllRqlPartsParamsFromMeta,
  getAttributeFilterRqlParts,
  getAttributesRqlParts,
  getFilterRqlParts,
  getGroupByClause,
  getRelationFilterRqlParts,
  getRelationsRqlParts,
  getRelationsSubjectObjectRqlParts,
  getRestrictionsClause,
  getSelectionClause,
  getSortJson,
  identity,
  parseEidList,
} from "../../utils/getParams";
import { initVariableKeyAssociation } from "../../utils/constants";

const personEntity: EntityRawSchema = {
  type: "Person",
  description: "A Person",
  final: false,
};

const taskEntity: EntityRawSchema = {
  type: "Task",
  description: "Something that must be done",
  final: false,
};

const stringEntity: EntityRawSchema = {
  type: "String",
  description: "",
  final: true,
};

const hasTaskRelation: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "has_task",
  description: "Something that the person must do",
  final: false,
  cardinality: "*+",
  subject: "Person",
  object: "Task",
  constraints: [],
  options: {},
};

const hasMainTaskRelation: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "has_main_task",
  description: "Something that the person must do",
  final: false,
  cardinality: "11",
  subject: "Person",
  object: "Task",
  constraints: [],
  options: {},
};

const friendRelation: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "friend",
  description: "A peron can be friend wit another person",
  final: false,
  cardinality: "**",
  subject: "Person",
  object: "Person",
  constraints: [],
  options: {},
};

const nameAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "name",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "String",
  constraints: [],
  default: null,
  options: {},
};

const surnameAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "surname",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "String",
  constraints: [],
  default: null,
  options: {},
};

const eidAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "eid",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "Int",
  constraints: [],
  default: null,
  options: {},
};

const baseEntities = [personEntity, taskEntity, stringEntity];

const emptySchema = new Schema({
  entities: baseEntities,
  relations_definitions: [],
});

const oneRelationSchema = new Schema({
  entities: baseEntities,
  relations_definitions: [hasMainTaskRelation],
});

const oneRelationMultipleCardinalitySchema = new Schema({
  entities: baseEntities,
  relations_definitions: [hasTaskRelation],
});

const multipleRelationsSchema = new Schema({
  entities: baseEntities,
  relations_definitions: [hasTaskRelation, hasMainTaskRelation],
});

const oneRelationAndAttributeSchema = new Schema({
  entities: baseEntities,
  relations_definitions: [hasMainTaskRelation, nameAttribute],
});

const nameEidSchema = new Schema({
  entities: [personEntity, stringEntity],
  relations_definitions: [eidAttribute, nameAttribute, surnameAttribute],
});

const symetricSchema = new Schema({
  entities: [personEntity],
  relations_definitions: [friendRelation],
});

describe("getFilterRqlParts", () => {
  const entity = nameEidSchema.getEntity("Person") as EntitySchema;
  it("should create filters from arrays", () => {
    expect(
      getFilterRqlParts(
        { relations: { eid: { ids: { include: [1] } } } },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject(["EXISTS(X eid IN (1))"]);
    expect(
      getFilterRqlParts(
        { relations: { eid: { ids: { include: ["1"] } } } },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject(["EXISTS(X eid IN (1))"]);
  });
  it("should create filters from strings", () => {
    expect(
      getFilterRqlParts(
        { relations: { name: { text: "test" } } },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject(["EXISTS(X name ~= '%test%')"]);
    expect(
      getFilterRqlParts(
        { relations: { name: { text: "test" }, surname: { text: "test2" } } },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "EXISTS(X name ~= '%test%')",
      "EXISTS(X surname ~= '%test2%')",
    ]);
  });
  it("should create custom RQL filters", () => {
    expect(
      getFilterRqlParts(
        { rql: ["custom rql query"] },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject(["custom rql query"]);
    expect(
      getFilterRqlParts(
        {
          rql: ["custom rql query 1", "custom rql query 2"],
        },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject(["custom rql query 1", "custom rql query 2"]);
  });
  it("should mix all types", () => {
    expect(
      getFilterRqlParts(
        {
          rql: ["custom rql query 1", "custom rql query 2"],
          relations: {
            eid: { ids: { include: [1, 2] } },
            name: { text: "test" },
          },
        },
        entity,
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "custom rql query 1",
      "custom rql query 2",
      "EXISTS(X eid IN (1, 2))",
      "EXISTS(X name ~= '%test%')",
    ]);
  });
});

describe("getRelationFilterRqlParts", () => {
  it("should create filters from include arrays", () => {
    expect(
      getRelationFilterRqlParts(
        "Relation_key",
        { ids: { include: [1] } },
        "MAIN_VAR",
        "relation_type",
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "EXISTS(MAIN_VAR relation_type FILTER_REL_1, FILTER_REL_1 eid IN(1))",
    ]);
  });
  it("should create filters from include arrays in reverse relation", () => {
    expect(
      getRelationFilterRqlParts(
        "reverse_Relation_key",
        { ids: { include: [1] } },
        "MAIN_VAR",
        "relation_type",
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "EXISTS(FILTER_REL_1 relation_type MAIN_VAR, FILTER_REL_1 eid IN(1))",
    ]);
  });
  it("should create filters from exclude arrays", () => {
    expect(
      getRelationFilterRqlParts(
        "Relation_key",
        { ids: { exclude: [1] } },
        "MAIN_VAR",
        "relation_type",
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "NOT EXISTS(MAIN_VAR relation_type FILTER_REL_1, FILTER_REL_1 eid IN(1))",
    ]);
  });
  it("should create filters from exclude arrays in reverse relation", () => {
    expect(
      getRelationFilterRqlParts(
        "reverse_Relation_key",
        { ids: { exclude: [1] } },
        "MAIN_VAR",
        "relation_type",
        initVariableKeyAssociation()
      )
    ).toMatchObject([
      "NOT EXISTS(FILTER_REL_1 relation_type MAIN_VAR, FILTER_REL_1 eid IN(1))",
    ]);
  });
});

describe("getAttributeFilterRqlParts", () => {
  it("should create filters from include arrays", () => {
    expect(
      getAttributeFilterRqlParts(
        "Relation_key",
        { ids: { include: [1] } },
        "MAIN_VAR"
      )
    ).toMatchObject(["EXISTS(MAIN_VAR Relation_key IN (1))"]);
  });
  it("should create filters from exclude arrays", () => {
    expect(
      getAttributeFilterRqlParts(
        "Relation_key",
        { ids: { exclude: [1] } },
        "MAIN_VAR"
      )
    ).toMatchObject(["NOT EXISTS(MAIN_VAR Relation_key IN (1))"]);
  });
  it("should create filters from strings", () => {
    expect(
      getAttributeFilterRqlParts("Relation_key", { text: "test" }, "MAIN_VAR")
    ).toBe("EXISTS(MAIN_VAR Relation_key ~= '%test%')");
  });
});

describe("getSortJson", () => {
  it("should detect invalid data", () => {
    expect(getSortJson("truc")).toBeNull();
    expect(getSortJson(`{"random": "thing"}`)).toBeNull();
  });
  it("should recover the object", () => {
    expect(
      getSortJson(`{"orderExpression": "test1", "restriction": "test2"}`)
    ).toMatchObject({ orderExpression: "test1", restriction: "test2" });
  });
});

describe("getAllRqlPartsParamsFromMeta", () => {
  const defaultResult = {
    includeObjectRelations: false,
    includeSubjectRelations: false,
    resolveRelations: undefined,
    knownRelation: undefined,
    sortData: undefined,
  };
  it("handles empty meta", () => {
    expect(getAllRqlPartsParamsFromMeta()).toEqual(defaultResult);
  });
  it("handles meta with all relations", () => {
    expect(getAllRqlPartsParamsFromMeta({ resolveRelations: "all" })).toEqual({
      ...defaultResult,
      includeSubjectRelations: true,
      includeObjectRelations: true,
    });
  });
  it("handles meta with only subject", () => {
    expect(
      getAllRqlPartsParamsFromMeta({ resolveRelations: "allSubject" })
    ).toEqual({
      ...defaultResult,
      includeSubjectRelations: true,
      includeObjectRelations: false,
    });
  });
  it("handles meta with custom set of relations", () => {
    expect(
      getAllRqlPartsParamsFromMeta({
        resolveRelations: { custom_relation: true },
      })
    ).toEqual({
      ...defaultResult,
      includeSubjectRelations: true,
      includeObjectRelations: true,
      resolveRelations: { custom_relation: true },
    });
  });
});

describe("getAllRqlParts", () => {
  const emptyEntity = emptySchema.getEntity("Person") as EntitySchema;
  const attributeAndrelationEntity = oneRelationAndAttributeSchema.getEntity(
    "Person"
  ) as EntitySchema;
  it("handles empty schema", () => {
    expect(getAllRqlParts(emptyEntity)).toEqual({
      selectionTransformers: [identity],
      selection: ["X"],
      groupBy: ["X"],
      restrictions: [],
      relationsTypes: ["X"],
      sortExpression: null,
    });
  });

  const expectedValue = {
    selectionTransformers: [identity, identity, identity],
    selection: ["X", "ATTR_1", "REL_1"],
    groupBy: ["X", "ATTR_1", "REL_1"],
    restrictions: ["X name ATTR_1", "X has_main_task REL_1?, REL_1 is Task"],
    relationsTypes: ["X", "name", "Person__has_main_task__Task"],
    sortExpression: null,
  };
  it("handles non empty schema with attributes and relations", () => {
    expect(
      getAllRqlParts(attributeAndrelationEntity, undefined, true, true)
    ).toEqual(expectedValue);
  });
  it("handles non empty schema with attributes and relations with simple attribute filter", () => {
    expect(
      getAllRqlParts(attributeAndrelationEntity, "name", true, true)
    ).toEqual({
      ...expectedValue,
      sortExpression: "ATTR_1",
    });
  });
  it("handles non empty schema with attributes and relations with complex attribute filter", () => {
    expect(
      getAllRqlParts(
        attributeAndrelationEntity,
        JSON.stringify({
          orderExpression: "custom order",
          restriction: "custom restrict",
        }),
        true,
        true
      )
    ).toEqual({
      ...expectedValue,
      restrictions: [
        "X name ATTR_1",
        "X has_main_task REL_1?, REL_1 is Task",
        "custom restrict",
      ],
      sortExpression: "custom order",
    });
  });
});

describe("getRelationsSubjectObjectRqlParts", () => {
  const entity = symetricSchema.getEntity("Person") as EntitySchema;
  const emptyObject = {
    selectionTransformers: [],
    selection: [],
    groupBy: [],
    restrictions: [],
    relationsTypes: [],
    sortVariable: null,
  };
  it("handles when both subject and object are not wanted", () => {
    expect(
      getRelationsSubjectObjectRqlParts(
        entity,
        initVariableKeyAssociation(),
        undefined,
        false,
        false
      )
    ).toEqual(emptyObject);
  });
  it("handles when only subject is wanted", () => {
    expect(
      getRelationsSubjectObjectRqlParts(
        entity,
        initVariableKeyAssociation(),
        undefined,
        true,
        false
      )
    ).toEqual(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    );
  });
  it("handles when only object is wanted", () => {
    expect(
      getRelationsSubjectObjectRqlParts(
        entity,
        initVariableKeyAssociation(),
        undefined,
        false,
        true
      )
    ).toEqual(
      getRelationsRqlParts(entity, "object", initVariableKeyAssociation())
    );
  });
  it("handles when both subject and object are wanted", () => {
    expect(
      getRelationsSubjectObjectRqlParts(
        entity,
        initVariableKeyAssociation(),
        undefined,
        true,
        true
      )
    ).toEqual({
      selectionTransformers: [parseEidList, parseEidList],
      selection: ["GROUP_CONCAT(REL_1)", "GROUP_CONCAT(REL_2)"],
      groupBy: [],
      restrictions: [
        "X friend REL_1?, REL_1 is Person",
        "REL_2? friend X, REL_2 is Person",
      ],
      relationsTypes: [
        "Person__friend__Person",
        "reverse_Person__friend__Person",
      ],
      sortVariable: null,
    });
  });
});

describe("getRelationsRqlParts", () => {
  it("should handle empty schema", () => {
    const entity = emptySchema.getEntity("Person") as EntitySchema;
    expect(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [],
      selection: [],
      groupBy: [],
      restrictions: [],
      sortVariable: null,
    });
  });
  it("should find one relation", () => {
    const entity = oneRelationSchema.getEntity("Person") as EntitySchema;
    expect(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [identity],
      selection: ["REL_1"],
      groupBy: ["REL_1"],
      restrictions: [`X has_main_task REL_1?, REL_1 is Task`],
      sortVariable: null,
    });
  });
  it("should handle relations of multiple cardinality", () => {
    const entity = oneRelationMultipleCardinalitySchema.getEntity(
      "Person"
    ) as EntitySchema;
    expect(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [parseEidList],
      selection: ["GROUP_CONCAT(REL_1)"],
      groupBy: [],
      restrictions: ["X has_task REL_1?, REL_1 is Task"],
      sortVariable: null,
    });
  });
  it("should find multiple relations", () => {
    const entity = multipleRelationsSchema.getEntity("Person") as EntitySchema;
    expect(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [parseEidList, identity],
      selection: ["GROUP_CONCAT(REL_1)", "REL_2"],
      groupBy: ["REL_2"],
      restrictions: [
        `X has_task REL_1?, REL_1 is Task`,
        `X has_main_task REL_2?, REL_2 is Task`,
      ],
      sortVariable: null,
    });
  });
  it("should ignore attributes", () => {
    const entity = oneRelationAndAttributeSchema.getEntity(
      "Person"
    ) as EntitySchema;
    expect(
      getRelationsRqlParts(entity, "subject", initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [identity],
      selection: ["REL_1"],
      groupBy: ["REL_1"],
      restrictions: [`X has_main_task REL_1?, REL_1 is Task`],
      sortVariable: null,
    });
  });

  describe("subject", () => {
    it("should handle known relation", () => {
      // FIXME use a better variable name than M
      const entity = oneRelationSchema.getEntity("Person") as EntitySchema;
      expect(
        getRelationsRqlParts(
          entity,
          "subject",
          initVariableKeyAssociation(),
          undefined,
          {
            relation: "Person__has_main_task__Task",
            value: 10,
          }
        )
      ).toMatchObject({
        selectionTransformers: [identity],
        selection: ["REL_1"],
        groupBy: ["REL_1"],
        restrictions: [
          `X has_main_task REL_1?, REL_1 is Task, M eid 10, X has_main_task M`,
        ],
        sortVariable: null,
      });
    });
  });
  describe("object", () => {
    it("should handle known relation", () => {
      const entity = oneRelationSchema.getEntity("Task") as EntitySchema;
      expect(
        getRelationsRqlParts(
          entity,
          "object",
          initVariableKeyAssociation(),
          undefined,
          {
            relation: "reverse_Person__has_main_task__Task",
            value: 10,
          }
        )
      ).toMatchObject({
        selectionTransformers: [identity],
        selection: ["REL_1"],
        groupBy: ["REL_1"],
        restrictions: [
          `REL_1? has_main_task X, REL_1 is Person, M eid 10, M has_main_task X`,
        ],
        sortVariable: null,
      });
    });
  });
  it("should handle sort", () => {
    const entity = oneRelationSchema.getEntity("Person") as EntitySchema;
    expect(
      getRelationsRqlParts(
        entity,
        "subject",
        initVariableKeyAssociation(),
        "Person__has_main_task__Task"
      )
    ).toMatchObject({
      selectionTransformers: [identity],
      selection: ["REL_1"],
      groupBy: ["REL_1"],
      restrictions: [`X has_main_task REL_1?, REL_1 is Task`],
      sortVariable: "REL_1",
    });
  });
  it("should allow selecting relations", () => {
    const entity = multipleRelationsSchema.getEntity("Person") as EntitySchema;
    expect(
      getRelationsRqlParts(
        entity,
        "subject",
        initVariableKeyAssociation(),
        undefined,
        undefined,
        { has_main_task: true }
      )
    ).toMatchObject({
      selectionTransformers: [identity],
      selection: ["REL_1"],
      groupBy: ["REL_1"],
      restrictions: [`X has_main_task REL_1?, REL_1 is Task`],
      sortVariable: null,
    });
  });
});

describe("getAttributesRqlParts", () => {
  const entity = oneRelationAndAttributeSchema.getEntity(
    "Person"
  ) as EntitySchema;
  const emptyEntity = emptySchema.getEntity("Person") as EntitySchema;
  const multipleAttributesEntity = nameEidSchema.getEntity(
    "Person"
  ) as EntitySchema;
  it("should handle empty schema", () => {
    expect(
      getAttributesRqlParts(emptyEntity, initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [identity],
      relationsTypes: ["X"],
      selection: ["X"],
      groupBy: ["X"],
      restrictions: [],
      sortVariable: null,
    });
  });
  it("should return attributes", () => {
    expect(
      getAttributesRqlParts(
        multipleAttributesEntity,
        initVariableKeyAssociation()
      )
    ).toMatchObject({
      selectionTransformers: [identity, identity, identity],
      relationsTypes: ["X", "name", "surname"],
      selection: ["X", "ATTR_1", "ATTR_2"],
      groupBy: ["X", "ATTR_1", "ATTR_2"],
      restrictions: ["X name ATTR_1", "X surname ATTR_2"],
      sortVariable: null,
    });
  });
  it("should return only attributes data", () => {
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation())
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: null,
    });
  });
  it("should handle sort attribute", () => {
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), "name")
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: "ATTR_1",
    });
  });
  it("should handle X, id and eid sort attribute the same way", () => {
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), "X")
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: "X",
    });
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), "id")
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: "X",
    });
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), "eid")
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: "X",
    });
  });
  it("should handle known attributes", () => {
    // FIXME should we create variable and groupby when relation is known ?
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), undefined, {
        relation: "name",
        value: "test",
      })
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name test"],
      sortVariable: null,
    });
  });
  it("should ignore invalid known attributes", () => {
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), undefined, {
        relation: "test3",
        value: "test",
      })
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: null,
    });
  });
  it("should ignore known X", () => {
    expect(
      getAttributesRqlParts(entity, initVariableKeyAssociation(), undefined, {
        relation: "X",
        value: "test",
      })
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: null,
    });
  });
  it("should allow selecting attributes", () => {
    expect(
      getAttributesRqlParts(
        multipleAttributesEntity,
        initVariableKeyAssociation(),
        undefined,
        undefined,
        { name: true }
      )
    ).toMatchObject({
      selectionTransformers: [identity, identity],
      relationsTypes: ["X", "name"],
      selection: ["X", "ATTR_1"],
      groupBy: ["X", "ATTR_1"],
      restrictions: ["X name ATTR_1"],
      sortVariable: null,
    });
  });
});

describe("parseEidList", () => {
  it("handles empty string", () => {
    expect(parseEidList("")).toEqual([]);
  });
  it("handles one valid int string", () => {
    expect(parseEidList("1")).toEqual([1]);
  });
  it("handles 2 valid int strings", () => {
    expect(parseEidList("1, 2")).toEqual([1, 2]);
  });
  it("handles several valid int strings", () => {
    expect(parseEidList("1, 2, 3, 546, 354324, 54642")).toEqual([
      1, 2, 3, 546, 354324, 54642,
    ]);
  });
  it("handles non-valid string", () => {
    expect(parseEidList("this is a test")).toEqual([]);
  });
  it("handles non int items in string", () => {
    expect(parseEidList("1, 2, 3, hello, 354324, 54642")).toEqual([
      1, 2, 3, 354324, 54642,
    ]);
  });
  it("handles missing comma", () => {
    expect(parseEidList("1, 2, 3 354324, 54642")).toEqual([1, 2, 3, 54642]);
  });
});

describe("getSelectionClause", () => {
  it("throws an error on empty selection", () => {
    expect(() => getSelectionClause([])).toThrowError("RQL selection is empty");
  });
  it("handles selection with one item", () => {
    expect(getSelectionClause(["A"])).toBe("Any A");
  });
  it("handles selection with multiple item", () => {
    expect(getSelectionClause(["A", "B", "C"])).toBe("Any A, B, C");
  });
});

describe("getRestrictionsClause", () => {
  it("handles empty restrictions without entity", () => {
    expect(getRestrictionsClause([])).toBe("");
  });
  it("handles empty restrictions with an entity", () => {
    expect(getRestrictionsClause([], "Task")).toBe("");
  });
  it("handles one restriction without entity", () => {
    expect(getRestrictionsClause(["A is Test"])).toBe("WHERE A is Test");
  });
  it("handles one restriction with an entity", () => {
    expect(getRestrictionsClause(["A is Test"], "Task")).toBe(
      "WHERE X is Task, A is Test"
    );
  });
  it("handles several restrictions without entity", () => {
    expect(getRestrictionsClause(["A is Test", "B eid 51"])).toBe(
      "WHERE A is Test, B eid 51"
    );
  });
  it("handles several restrictions with an entity", () => {
    expect(getRestrictionsClause(["A is Test", "B eid 51"], "Task")).toBe(
      "WHERE X is Task, A is Test, B eid 51"
    );
  });
});

describe("getGroupByClause", () => {
  it("handles empty array", () => {
    expect(getGroupByClause([])).toBe("");
  });
  it("handles one item", () => {
    expect(getGroupByClause(["test"])).toBe("GROUPBY test");
  });
  it("handles several items item", () => {
    expect(getGroupByClause(["test1", "test2"])).toBe("GROUPBY test1, test2");
  });
});
