import {
  EntityRawSchema,
  EntityRawSchemaArray,
  EntitySchema,
  RelationDefinitionRawSchema,
  Schema,
  Transaction,
  TransactionQuery,
  getLooseSchema,
} from "@cubicweb/client";
import {
  generateRqlUpdateParts,
  generateRqlUpdateRelationsParts,
  generateRqlUpdateRelationsPartsFromInput,
  makeUpdateRequest,
} from "../../utils/updateParams";

describe("generateRqlUpdateRelationsParts", () => {
  it("should return empty data", () => {
    expect(
      generateRqlUpdateRelationsParts("Relation", [], "subject", "set")
    ).toEqual([]);
  });
  describe("subject", () => {
    describe("set", () => {
      it("should find new items", () => {
        expect(
          generateRqlUpdateRelationsParts("Relation", [2], "subject", "set")
        ).toEqual([["X Relation SUB_SET_REL0", "SUB_SET_REL0 eid 2"]]);
      });
      describe("delete", () => {
        it("should find new items", () => {
          expect(
            generateRqlUpdateRelationsParts(
              "Relation",
              [2],
              "subject",
              "delete"
            )
          ).toEqual([["X Relation SUB_DEL_REL0", "SUB_DEL_REL0 eid 2"]]);
        });
      });
    });
    describe("object", () => {
      describe("set", () => {
        it("should find new items", () => {
          expect(
            generateRqlUpdateRelationsParts("Relation", [2], "object", "set")
          ).toEqual([["OBJ_SET_REL0 Relation X", "OBJ_SET_REL0 eid 2"]]);
        });
        describe("delete", () => {
          it("should find new items", () => {
            expect(
              generateRqlUpdateRelationsParts(
                "Relation",
                [2],
                "object",
                "delete"
              )
            ).toEqual([["OBJ_DEL_REL0 Relation X", "OBJ_DEL_REL0 eid 2"]]);
          });
        });
      });
    });
  });
});

describe("generateRqlUpdateRelationsPartsFromInput", () => {
  it("handles empty data", () => {
    expect(
      generateRqlUpdateRelationsPartsFromInput("relation", undefined, "subject")
    ).toEqual({
      relationsToDelete: [],
      relationsToSet: [],
    });
  });
  it("handles single id value", () => {
    expect(
      generateRqlUpdateRelationsPartsFromInput("relation", 5, "subject")
    ).toEqual({
      relationsToDelete: [],
      relationsToSet: [["X relation SUB_SET_REL0", "SUB_SET_REL0 eid 5"]],
    });
  });
  it("handles newly selected values", () => {
    expect(
      generateRqlUpdateRelationsPartsFromInput(
        "relation",
        {
          id: "temp_input",
          selected: [5, 10],
        },
        "subject"
      )
    ).toEqual({
      relationsToDelete: [],
      relationsToSet: [
        ["X relation SUB_SET_REL0", "SUB_SET_REL0 eid 5"],
        ["X relation SUB_SET_REL1", "SUB_SET_REL1 eid 10"],
      ],
    });
  });
  it("handles values to be removed", () => {
    expect(
      generateRqlUpdateRelationsPartsFromInput(
        "relation",
        {
          id: "temp_input",
          removed: [5, 10],
        },
        "subject"
      )
    ).toEqual({
      relationsToDelete: [
        ["X relation SUB_DEL_REL0", "SUB_DEL_REL0 eid 5"],
        ["X relation SUB_DEL_REL1", "SUB_DEL_REL1 eid 10"],
      ],
      relationsToSet: [],
    });
  });
});

const personEntity: EntityRawSchema = {
  type: "Person",
  description: "A Person",
  final: false,
};

const taskEntity: EntityRawSchema = {
  type: "Task",
  description: "Something that must be done",
  final: false,
};
const stringEntity: EntityRawSchema = {
  type: "String",
  description: "",
  final: true,
};

const hasMainTaskRelation: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "has_main_task",
  description: "Something that the person must do",
  final: false,
  cardinality: "11",
  subject: "Person",
  object: "Task",
  constraints: [],
  options: {},
};

const nameAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "name",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "String",
  constraints: [],
  default: null,
  options: {},
};

const baseEntities = [personEntity, taskEntity, stringEntity];

const emptySchema = new Schema({
  entities: baseEntities,
  relations_definitions: [],
});

const oneRelationAndAttributeSchema = new Schema({
  entities: baseEntities,
  relations_definitions: [hasMainTaskRelation, nameAttribute],
});

describe("generateRqlUpdateParts", () => {
  const entity = oneRelationAndAttributeSchema.getEntity(
    "Person"
  ) as EntitySchema;
  describe("handles empty data", () => {
    const emptyResult = {
      relationsToDelete: [],
      relationsToSet: [],
      attributeUpdates: [],
      attributesValues: {},
    };
    const emptyEntity = emptySchema.getEntity("Person") as EntitySchema;
    it("handles empty changed data", () => {
      expect(generateRqlUpdateParts(entity, {})).toEqual(emptyResult);
    });
    it("handles empty schema", () => {
      expect(generateRqlUpdateParts(emptyEntity, { test: 5 })).toEqual(
        emptyResult
      );
    });
  });
  it("handles attributes and relations", () => {
    expect(
      generateRqlUpdateParts(entity, {
        name: "plop",
        Person__has_main_task__Task: 5,
      })
    ).toEqual({
      attributeUpdates: ["X name %(name)s"],
      attributesValues: {
        name: "plop",
      },
      relationsToDelete: [],
      relationsToSet: [["X has_main_task SUB_SET_REL0", "SUB_SET_REL0 eid 5"]],
    });
  });
  it("ignores unknown relations", () => {
    expect(
      generateRqlUpdateParts(entity, {
        name: "plop",
        Person__has_main_task__Task: 5,
        invalid_relation: 7,
        invalid_attribute: "test",
      })
    ).toEqual({
      attributeUpdates: ["X name %(name)s"],
      attributesValues: {
        name: "plop",
      },
      relationsToDelete: [],
      relationsToSet: [["X has_main_task SUB_SET_REL0", "SUB_SET_REL0 eid 5"]],
    });
  });
});

describe("makeUpdateRequest", () => {
  describe("handles empty values", () => {
    it("handles empty data", () => {
      const transaction = new Transaction();
      expect(
        makeUpdateRequest(
          getLooseSchema(oneRelationAndAttributeSchema),
          "Person",
          {
            id: 5,
            data: {},
            previousData: {},
          },
          transaction
        )
      ).toEqual([]);
      expect(transaction.queries).toEqual([]);
    });
    it("handles empty schema", () => {
      const transaction = new Transaction();
      expect(
        makeUpdateRequest(
          getLooseSchema(emptySchema),
          "Person",
          {
            id: 5,
            data: { test: 5 },
            previousData: { test: 10 },
          },
          transaction
        )
      ).toEqual([]);
      expect(transaction.queries).toEqual([]);
    });
  });
  it("handles attribute change", () => {
    const transaction = new Transaction();
    expect(
      makeUpdateRequest(
        getLooseSchema(oneRelationAndAttributeSchema),
        "Person",
        {
          id: 5,
          data: { name: "new" },
          previousData: { name: "old" },
        },
        transaction
      )
    ).toEqual([
      {
        column: 0,
        queryIndex: 0,
        row: 0,
      },
    ]);
    expect(transaction.queries).toEqual([
      new TransactionQuery(
        "SET X name %(name)s WHERE X is Person, X eid 5",
        { name: "new" },
        0
      ),
    ]);
  });

  it("handles relation change", () => {
    const transaction = new Transaction();
    expect(
      makeUpdateRequest(
        getLooseSchema(oneRelationAndAttributeSchema),
        "Person",
        {
          id: 5,
          data: { Person__has_main_task__Task: 5 },
          previousData: { Person__has_main_task__Task: 10 },
        },
        transaction
      )
    ).toEqual([
      {
        column: 0,
        queryIndex: 0,
        row: 0,
      },
    ]);
    expect(transaction.queries).toEqual([
      new TransactionQuery(
        "SET X has_main_task SUB_SET_REL0 WHERE SUB_SET_REL0 eid 5, X eid 5",
        {},
        0
      ),
    ]);
  });
  // TODO implement and test handling of query refs when creating an entity in the update request
});
