import {
  EntityRawSchemaArray,
  RelationDefinitionRawSchemaArray,
  RelationDefinitionSchema,
  EntitySchema,
  EntityRole,
} from "@cubicweb/client";
import {
  ENTITY_TYPES_METAMODEL,
  RELATIONS_TO_IGNORE,
  RELATION_OBJECTS_TO_IGNORE,
  RELATIONS_AND_ATTRIBUTES_METAMODEL,
  RELATIONS_AND_ATTRIBUTES_METAMODEL_INFOBOX,
} from "../CubicWebMetamodel";
import { getRelationKey } from "../relations";

/**
 * Ignores non-final entities
 *
 * @param entities The original list of entities
 * @returns List of non-final entities
 */
export const getNonFinalEntities = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  ERawSchema extends E[number]
>(
  entities: EntitySchema<E, R, ERawSchema>[]
) => entities.filter((e) => !e.final);

/**
 * Checks if the given entity is from cubicweb's core
 *
 * @param entity The entity to check
 * @returns True if the entity is from CubicWeb's core, false otherwise
 */
export const isCoreEntity = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  ERawSchema extends E[number]
>(
  entity: EntitySchema<E, R, ERawSchema>
) => ENTITY_TYPES_METAMODEL.includes(entity.type);

/**
 * Ignores relations which should not be shown in react admin.
 * Those also include relations with an incompatible object type.
 *
 * @param relations The original list of relations
 * @returns Filtered list of relations
 */
export const getCompatibleRelations = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  RDef extends R[number]
>(
  relations: RelationDefinitionSchema<E, R, RDef>[]
) =>
  relations.filter(
    (e) =>
      !RELATIONS_TO_IGNORE.includes(e.type) &&
      !RELATION_OBJECTS_TO_IGNORE.includes(e.object.type)
  );

/**
 * Checks if the given relations is from cubicweb's core and can be edited
 *
 * @param relation The relation to check
 * @returns True if the relation is from CubicWeb's core and is editable, false otherwise
 */
export const isCoreEditableRelation = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  RDef extends R[number]
>(
  relation: RelationDefinitionSchema<E, R, RDef>
) => RELATIONS_AND_ATTRIBUTES_METAMODEL.includes(relation.type);

/**
 * Checks if the given relations is from cubicweb's core and cannot be edited
 *
 * @param relation The relation to check
 * @returns True if the relation is from CubicWeb's core and is read only, false otherwise
 */
export const isCoreReadOnlyRelation = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  RDef extends R[number]
>(
  relation: RelationDefinitionSchema<E, R, RDef>
) => RELATIONS_AND_ATTRIBUTES_METAMODEL_INFOBOX.includes(relation.type);

/**
 * Creates a list of attributes names for the given entity.
 * It only returns attributes from cubes and ignores the
 * CubicWeb Metamodel.
 *
 * @param entitySchema The entity schema abstraction class
 * @returns A list of attributes names as string
 */
export const getAttributesNames = (entitySchema: EntitySchema) => {
  const cubeAttributesNames = getEntitySchemaNames(entitySchema, "attributes");
  return ["X", ...cubeAttributesNames];
};

/**
 * Retrieves the list of relation names of the given type.
 *
 * @param entitySchema The entity schema abstraction class
 * @param type The type of relations to fetch
 * @returns A list of names for the given relation type
 */
export const getEntitySchemaNames = (
  entitySchema: EntitySchema,
  type: "attributes" | EntityRole
): string[] => {
  let relations: RelationDefinitionSchema[] = [];
  switch (type) {
    case "attributes":
      relations = entitySchema.attributes;
      break;
    case "subject":
    case "object":
      relations = entitySchema.relationDefinitions[type];
      break;
  }
  return getCompatibleRelations(relations).map((a) => {
    if (a.final) {
      return a.type;
    } else {
      return getRelationKey(a, type === "attributes" ? "subject" : type);
    }
  });
};
